# ANTS-Hub - Copyright (C) Entr'ouvert

import logging

import requests
from django.db.models import F, Q
from django.db.transaction import atomic

from ants_hub.data.models import Config, RendezVous
from ants_hub.timezone import localtime, now

logger = logging.getLogger('ants_hub.api.ants')


class ANTSError(Exception):
    pass


def get_api_optimisation_auth_token():
    return Config.get(Config.REQUEST_TO_ANTS_V2_AUTH_TOKEN)


def get_api_optimisation_url():
    return Config.get(
        Config.REQUEST_TO_ANTS_V2_BASE_URL, 'https://api-coordination.rendezvouspasseport.ants.gouv.fr/api/'
    )


def push_rdv(rdv):
    # swagger : https://api-coordination.rendezvouspasseport.ants.gouv.fr/docs#/Appointments%20-%20%C3%A9diteurs/add_appointment_api_appointments_post
    auth_token = get_api_optimisation_auth_token()
    if auth_token is None:
        raise ANTSError('REQUEST_TO_ANTS_V2_AUTH_TOKEN not configured')
    base_url = get_api_optimisation_url()
    try:
        if rdv.canceled is None:
            params = [
                ('application_id', rdv.identifiant_predemande),
                ('management_url', rdv.gestion_url),
                ('meeting_point', rdv.lieu.nom),
                ('meeting_point_id', str(rdv.lieu.id)),
                ('appointment_date', localtime(rdv.date).strftime('%Y-%m-%d %H:%M:%S')),
            ]
            response = requests.post(
                base_url + 'appointments',
                params=params,
                headers={'x-rdv-opt-auth-token': auth_token},
                timeout=10,
            )
            response.raise_for_status()
            return response.json()['success'] == 'true'
        else:
            params = [
                ('application_id', rdv.identifiant_predemande),
                ('meeting_point', rdv.lieu.nom),
                ('meeting_point_id', str(rdv.lieu.id)),
                ('appointment_date', localtime(rdv.date).strftime('%Y-%m-%d %H:%M:%S')),
            ]
            response = requests.delete(
                base_url + 'appointments',
                params=params,
                headers={'x-rdv-opt-auth-token': auth_token},
                timeout=10,
            )
            response.raise_for_status()
            return response.json()['rowcount'] == 1
    except (requests.RequestException, ValueError, KeyError, TypeError) as e:
        raise ANTSError(str(e))


def get_rdvs_to_upload():
    return RendezVous.objects.filter(
        Q(last_upload__isnull=True) | Q(last_upload__lt=F('last_update'))
    ).select_related('lieu')


def upload_rdvs():
    # get never uploaded rdvs and rdv uploaded which changed (cancelled)
    with atomic():
        rdvs = get_rdvs_to_upload().select_for_update(of=('self',)).distinct()
        start = now()
        pushed = set()
        for rdv in rdvs:
            try:
                if push_rdv(rdv):
                    pushed.add(rdv)
            except ANTSError as e:
                logger.warning('unable to push rdv(%s) of lieu %s: %r', rdv, rdv.lieu, e)
        rdvs.filter(pk__in=[rdv.pk for rdv in pushed]).update(last_upload=start)
        for rdv in pushed:
            logger.info('pushed rdv(%s) of lieu %s', rdv, rdv.lieu)


def get_status_of_predemandes(identifiant_predemandes: list):
    auth_token = get_api_optimisation_auth_token()
    base_url = get_api_optimisation_url()

    params = [
        ('application_ids', identifiant_predemande) for identifiant_predemande in identifiant_predemandes
    ]

    try:
        response = requests.get(
            base_url + 'status',
            params=params,
            headers={'x-rdv-opt-auth-token': auth_token},
            timeout=10,
        )
        response.raise_for_status()
        data = response.json()
        if not isinstance(data, dict):
            raise ValueError(f'reponse is not a dict: {data!r}')
    except (requests.RequestException, ValueError, TypeError) as e:
        raise ANTSError(repr(e))

    valid = True
    msg = []
    appointments = []
    for identifiant_predemande in identifiant_predemandes:
        if identifiant_predemande not in data:
            msg.append(f'Prédemande "{identifiant_predemande}" inconnue.')
            valid = False
            continue
        predemande_state = data[identifiant_predemande]
        if not isinstance(predemande_state, dict):
            raise ANTSError(f'application_id state is not a dict: {data!r}')
        if predemande_state.get('status') != 'validated':
            msg.append(f'Prédemande "{identifiant_predemande}" inconnue, expirée ou déjà consommée.')
            valid = False
            continue
        if predemande_state.get('appointments', []):
            appointments.extend(
                [
                    {'identifiant_predemande': identifiant_predemande, **appointment}
                    for appointment in predemande_state['appointments']
                ]
            )
            msg.append(f'Prédemande "{identifiant_predemande}" déjà liée à un ou des rendez-vous.')
            valid = False
            continue
    return valid, ' '.join(msg), data, appointments
